﻿using Org.BouncyCastle.Crypto.Digests;
using Org.BouncyCastle.Crypto.Generators;
using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Math.EC;
using Org.BouncyCastle.Security;
using System.Collections.Concurrent;

namespace YS.SM2
{
    /// <summary>
    /// SM2代理
    /// </summary>
    public class SM2Provider
    {
        private static readonly ConcurrentDictionary<string, string> sm2_param = new ConcurrentDictionary<string, string>
        {
            ["Ecc_Q"] = "FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFF",
            ["Ecc_A"] = "FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF00000000FFFFFFFFFFFFFFFC",
            ["Ecc_B"] = "28E9FA9E9D9F5E344D5A9E4BCF6509A7F39789F515AB8F92DDBCBD414D940E93",
            ["Ecc_N"] = "FFFFFFFEFFFFFFFFFFFFFFFFFFFFFFFF7203DF6B21C6052B53BBF40939D54123",
            ["Ecc_X"] = "32C4AE2C1F1981195F9904466A39C9948FE30BBFF2660BE1715A4589334C74C7",
            ["Ecc_Y"] = "BC3736A2F4F6779C59BDCEE36B692153D0A9877CC62A474002DF32E52139F0A0",
        };
        /// <summary>
        /// Ecc_Q
        /// </summary>
        private BigInteger Ecc_Q
        {
            get
            {
                return new BigInteger(sm2_param.GetValueOrDefault("Ecc_Q"), 16);
            }
        }
        /// <summary>
        /// Ecc_A
        /// </summary>
        private BigInteger Ecc_A
        {
            get
            {
                return new BigInteger(sm2_param.GetValueOrDefault("Ecc_A"), 16);
            }
        }
        /// <summary>
        /// Ecc_B
        /// </summary>
        private BigInteger Ecc_B
        {
            get
            {
                return new BigInteger(sm2_param.GetValueOrDefault("Ecc_B"), 16);
            }
        }
        /// <summary>
        /// Ecc_N
        /// </summary>
        private BigInteger Ecc_N
        {
            get
            {
                return new BigInteger(sm2_param.GetValueOrDefault("Ecc_N"), 16);
            }
        }
        /// <summary>
        /// Ecc_X
        /// </summary>
        private BigInteger Ecc_X
        {
            get
            {
                return new BigInteger(sm2_param.GetValueOrDefault("Ecc_X"), 16);
            }
        }
        /// <summary>
        /// Ecc_Y
        /// </summary>
        private BigInteger Ecc_Y
        {
            get
            {
                return new BigInteger(sm2_param.GetValueOrDefault("Ecc_Y"), 16);
            }
        }
        /// <summary>
        /// Ecc_Curve
        /// </summary>
        public ECCurve Ecc_Curve { get; }
        /// <summary>
        /// Ecc_Key_Pair_Generator
        /// </summary>
        public ECKeyPairGenerator Ecc_Key_Pair_Generator { get; }
        private static SM2Provider _instance;
        private static readonly object _lock = new object();
        /// <summary>
        /// 构造SM2代理
        /// </summary>
        private SM2Provider()
        {
            Ecc_Curve = new FpCurve(Ecc_Q, Ecc_A, Ecc_B, null, null);
            var ecc_point_g = Ecc_Curve.CreatePoint(Ecc_X, Ecc_Y);
            var ecc_bc_spec = new ECDomainParameters(Ecc_Curve, ecc_point_g, Ecc_N);
            var ecc_ecgenparam = new ECKeyGenerationParameters(ecc_bc_spec, new SecureRandom());
            Ecc_Key_Pair_Generator = new ECKeyPairGenerator();
            Ecc_Key_Pair_Generator.Init(ecc_ecgenparam);
        }
        /// <summary>
        /// 单例实例
        /// </summary>
        public static SM2Provider Instance
        {
            get
            {
                if (_instance == null)
                {
                    lock (_lock)
                    {
                        if (_instance == null)
                            _instance = new SM2Provider();
                    }
                }
                return _instance;
            }
        }
        /// <summary>
        /// 获取SM2Z
        /// </summary>
        /// <param name="userId">userId</param>
        /// <param name="userKey">userKey</param>
        /// <returns></returns>
        public virtual byte[] Sm2GetZ(byte[] userId, ECPoint userKey)
        {
            var sm3 = new SM3Digest();
            byte[] p;
            var len = userId.Length * 8;
            sm3.Update((byte)(len >> 8 & 0x00ff));
            sm3.Update((byte)(len & 0x00ff));
            sm3.BlockUpdate(userId, 0, userId.Length);
            p = Ecc_A.ToByteArray();
            sm3.BlockUpdate(p, 0, p.Length);
            p = Ecc_B.ToByteArray();
            sm3.BlockUpdate(p, 0, p.Length);
            p = Ecc_X.ToByteArray();
            sm3.BlockUpdate(p, 0, p.Length);
            p = Ecc_Y.ToByteArray();
            sm3.BlockUpdate(p, 0, p.Length);
            p = userKey.GetEncoded();
            sm3.BlockUpdate(p, 0, p.Length);
            p = userKey.GetEncoded();
            sm3.BlockUpdate(p, 0, p.Length);
            var md = new byte[sm3.GetDigestSize()];
            sm3.DoFinal(md, 0);
            return md;
        }
    }
}