﻿using Org.BouncyCastle.Crypto.Parameters;
using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Utilities.Encoders;
using System.Text;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Security;
using Org.BouncyCastle.Asn1.GM;
using Org.BouncyCastle.Crypto.Generators;

namespace YS.SM2
{
    /// <summary>
    /// SM2加解密
    /// </summary>
    public class SM2Utils
    {
        private readonly ECDomainParameters ecParams;
        private readonly SecureRandom secureRandom;
        /// <summary>
        /// SM2
        /// </summary>
        public SM2Utils()
        {
            var curve = GMNamedCurves.GetByName("sm2p256v1");
            ecParams = new ECDomainParameters(curve.Curve, curve.G, curve.N, curve.H);
            secureRandom = new SecureRandom();
        }
        /// <summary>
        /// 生成密码
        /// </summary>
        /// <returns></returns>
        public AsymmetricCipherKeyPair GenerateKeyPair()
        {
            ECKeyPairGenerator generator = new ECKeyPairGenerator();
            KeyGenerationParameters keyGenParam = new ECKeyGenerationParameters(ecParams, secureRandom);
            generator.Init(keyGenParam);
            return generator.GenerateKeyPair();
        }
        /// <summary>
        /// 加密
        /// </summary>
        /// <param name="data">需要加密的字节</param>
        /// <param name="publicKey">公钥</param>
        /// <returns></returns>
        /// <exception cref="ArgumentNullException"></exception>
        public byte[] Encrypt(byte[] data, string publicKey)
        {
            if (data == null)
                throw new ArgumentNullException(nameof(data));
            if (publicKey == null)
                throw new ArgumentNullException(nameof(publicKey));
            var pub= ConvertHexToPublicKey(publicKey);
            SM2Engine sm2Engine = new SM2Engine();
            sm2Engine.Init(true, new ParametersWithRandom(pub, secureRandom));
            return sm2Engine.ProcessBlock(data, 0, data.Length);
        }
        /// <summary>
        /// 解密
        /// </summary>
        /// <param name="encryptedData">需要解密的字节</param>
        /// <param name="privateKey">私钥</param>
        /// <returns></returns>
        /// <exception cref="ArgumentNullException"></exception>
        public byte[] Decrypt(byte[] encryptedData, string privateKey)
        {
            if (encryptedData == null)
                throw new ArgumentNullException(nameof(encryptedData));
            if (privateKey == null)
                throw new ArgumentNullException(nameof(privateKey));
            var pri = ConvertHexToPrivateKey(privateKey);
            SM2Engine sm2Engine = new SM2Engine();
            sm2Engine.Init(false, pri);
            return sm2Engine.ProcessBlock(encryptedData, 0, encryptedData.Length);
        }
        /// <summary>
        /// 字符串加密
        /// </summary>
        /// <param name="plainText">需要加密的字符串</param>
        /// <param name="publicKeyHex">公钥</param>
        /// <returns></returns>
        /// <exception cref="ArgumentNullException"></exception>
        public string Encrypt(string plainText, string publicKeyHex)
        {
            if (plainText == null)
                throw new ArgumentNullException(nameof(plainText));
            if (publicKeyHex == null)
                throw new ArgumentNullException(nameof(publicKeyHex));
            byte[] plainBytes = Encoding.UTF8.GetBytes(plainText);
            byte[] encryptedBytes = Encrypt(plainBytes, publicKeyHex);
            return Convert.ToBase64String(encryptedBytes);
        }
        /// <summary>
        /// 字符串解密
        /// </summary>
        /// <param name="cipherText">需要解密的字符串</param>
        /// <param name="privateKeyHex">私钥</param>
        /// <returns></returns>
        /// <exception cref="ArgumentNullException"></exception>
        public string Decrypt(string cipherText, string privateKeyHex)
        {
            if (cipherText == null)
                throw new ArgumentNullException(nameof(cipherText));
            if (privateKeyHex == null)
                throw new ArgumentNullException(nameof(privateKeyHex));
            byte[] cipherBytes = Convert.FromBase64String(cipherText);
            byte[] decryptedBytes = Decrypt(cipherBytes, privateKeyHex);
            return Encoding.UTF8.GetString(decryptedBytes);
        }
        /// <summary>
        /// 公钥转换
        /// </summary>
        /// <param name="publicKeyHex">公钥</param>
        /// <returns></returns>
        private ECPublicKeyParameters ConvertHexToPublicKey(string publicKeyHex)
        {
            var q = ecParams.Curve.DecodePoint(Hex.Decode(publicKeyHex));
            return new ECPublicKeyParameters(q, ecParams);
        }
        /// <summary>
        /// 私钥转换
        /// </summary>
        /// <param name="privateKeyHex">私钥</param>
        /// <returns></returns>
        private ECPrivateKeyParameters ConvertHexToPrivateKey(string privateKeyHex)
        {
            var d = new BigInteger(privateKeyHex, 16);
            return new ECPrivateKeyParameters(d, ecParams);
        }
    }
}
